/*
 * Copyright (c) 2004 Cat's Eye Technologies.
 * Copyright (c) 1980, 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file was derived from:
 *
 * @(#) Copyright (c) 1980, 1992, 1993 The Regents of the University of California.  All rights reserved.
 * @(#)script.c	8.1 (Berkeley) 6/6/93
 * $FreeBSD: src/usr.bin/script/script.c,v 1.11.2.1 2000/07/20 10:35:21 kris Exp $
 * $DragonFly: src/usr.bin/script/script.c,v 1.3 2003/10/04 20:36:50 hmp Exp $
 *
 * $Id: brace.c 11 2004-09-21 02:08:33Z catseye $
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <libutil.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <sysexits.h>

/* #include <curses.h> */

/* constants */
#define BUFSIZE		4096
#define ESO_ESC		(char)5

/* should be enums */
#define ST_CAT		0
#define ST_CMDMAJ	1
#define ST_CMDMIN	2
#define ST_ARGCHAR	3
#define ST_ARGINT	4

/* GLOBALS */

char		*this_prog;
char		*target_prog;
int		 master, slave;
int		 child;
struct termios	 tt;
int		 rand_in = 0;
int		 rand_out = 0;
int		 state = ST_CAT;

/* prototypes */

static void	done(int) __dead2;
static void	usage(void);
static void	process_input(int, char *, size_t);
static void	process_output(int, char *, size_t);
static void	strrandcase(char *, size_t);

int
main(int argc, char **argv)
{
	int cc;
	struct termios rtt, stt;
	struct winsize win;
	int ch, n;
	char obuf[BUFSIZ];
	char ibuf[BUFSIZ];
	fd_set rfd;
	int lastc = 1000;
	int die, e, pid;
	union wait status;

	this_prog = argv[0];

	/*
	 * Get command-line arguments.
	 */
	while ((ch = getopt(argc, argv, "r:")) != -1)
		switch(ch) {
		case 'r':
			rand_in = (strchr(optarg, 'i') != NULL);
			rand_out = (strchr(optarg, 'o') != NULL);
			if (!(rand_in || rand_out))
				usage();
			srandomdev();
			break;
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc == 0)
		usage();

	target_prog = argv[0];

	/*
	 * Create the pty.  Give it the same terminal attributes
	 * as the terminal we're currently using.
	 */
	tcgetattr(STDIN_FILENO, &tt);
	ioctl(STDIN_FILENO, TIOCGWINSZ, &win);
	if (openpty(&master, &slave, NULL, &tt, &win) == -1)
		err(EX_OSERR, "openpty");

	/*
	 * Make stdin 'raw' and stop it from echoing,
	 * otherwise we'll get doubled output.
	 */
	rtt = tt;
	cfmakeraw(&rtt);
	rtt.c_lflag &= ~ECHO;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &rtt);

	/*
	initscr();
	cbreak();
	noecho();
	nl();
	*/

	/*
	 * Fork off a child process to handle the execution
	 * of the target program.
	 */
	child = fork();
	if (child < 0) {
		/*
		 * Oh no!  Something went wrong!
		 */
		warn("fork");
		done(EX_OSERR);
	}
	if (child == 0) {
		/*
		 * OK, looks like *we* are the child.
		 */
		close(master);
		login_tty(slave);
		execvp(argv[0], argv);
		/*
		 * If we got this far, we obviously failed to
		 * execute the program.
		 */
		warn("%s", argv[0]);
		kill(0, SIGTERM);
		done(EX_OSERR);
	}

	/*
	 * Otherwise, we're the parent.
	 * Loop, selecting on stdin and 'master' (the program.)
	 */
	FD_ZERO(&rfd);
	for (;;) {
		FD_SET(master, &rfd);
		FD_SET(STDIN_FILENO, &rfd);
		n = select(master + 1, &rfd, 0, 0, NULL);
		if (n < 0 && errno != EINTR)
			break;
		if (n > 0 && FD_ISSET(STDIN_FILENO, &rfd)) {
			cc = read(STDIN_FILENO, ibuf, sizeof(ibuf));
			if (cc <= 0)
				break;
			if (cc > 0) {
				/* Send keystrokes to program. */
				process_input(master, ibuf, cc);
			}
		}
		if (n > 0 && FD_ISSET(master, &rfd)) {
			cc = read(master, obuf, sizeof(obuf));
			if (cc <= 0)
				break;
			/* send program output to stdout */
			process_output(STDOUT_FILENO, obuf, cc);
		}
	}

	/*
	if (cc < 0)
		err(EX_IOERR, "read/write");
	*/

	/*
	 * Clean up and exit.
	 */
	die = e = 0;
	while ((pid = wait3((int *)&status, WNOHANG, 0)) > 0)
	        if (pid == child) {
			die = 1;
			if (WIFEXITED(status))
				e = WEXITSTATUS(status);
			else if (WIFSIGNALED(status))
				e = WTERMSIG(status);
			else /* can't happen */
				e = EX_OSERR;
		}

	if (die)
		done(e);

	done(0);
}

void
process_input(int fd, char *buf, size_t len)
{
	if (rand_in)
		strrandcase(buf, len);
	write(fd, buf, len);
}

void
process_output(int fd, char *buf, size_t len)
{
	size_t i = 0, start = 0;

	while (i < len) {
		switch (state) {
		case ST_CAT:
			if (rand_out)
				strrandcase(buf, len);
			write(fd, buf, len);
			i = len;
			break;
		case ST_CMDMAJ:
			i++;
			state = ST_CMDMIN;
			start = i;
			break;
		case ST_CMDMIN:
			i++;
			state = ST_CAT;
			start = i;
			break;
		}
	}
}

/*
 * Handle throughput, including the finite state machine.
 * In ST_CAT we simply pass the buffer through unmodified
 * (except for certain control sequences which we must intercept.)
 *
void
thruput(int dir, int fd1, int fd2, char *s, size_t len)
{
	char b;

	while (i < len) {
		switch (state) {
		case ST_CAT:
			 *
			 * Find everything until the next magic character,
			 * and spit it out in one chunk for efficiency.
			 *
			for (i = start; i < len && !iscntrl(s[i]); i++)
				;
		
			if (i != start) {
				if (randcase)
					strrandcase(s + start, i - start);
				write2(fd1, fd2, s + start, i - start);
			}
		
			if (i < len) {
				if (s[i] == '\n' || s[i] == '\r') {
					write2(fd1, fd2, "\r\n", 2);
				} else if (s[i] == ESO_ESC) {
					if (dir == INPUT) {
						write2(fd1, -1, "\005", 1);
					} else {
						write2(fd1, fd2, "*ESO*", 5);
						state = ST_CMDMAJ;
					}
				} else {
					write2(fd1, fd2, &s[i], 1);
				}
				i++;
			}
			start = i;
			break;
	}
}
*/
	

/*
 * cOnvERt a BUfFER tO raNdoM CAse.
 */
void
strrandcase(char *s, size_t len)
{
	size_t i;

	for(i = 0; i < len; i++)
		s[i] = random() & 0x01 ? toupper(s[i]) : tolower(s[i]);
}

static void
usage(void)
{
	fprintf(stderr,
	    "usage: %s [-r i|o|io] command\n", this_prog);
	exit(EX_USAGE);
}

void
done(int eno)
{
	/*
	 * Reset the terminal back to its original state.
	 */
	/*
	endwin();
	*/
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &tt);
	close(master);
	exit(eno);
}
